@@ -29,7 +29,6 @@ from .key_binding import InputProcessor
 from .enums import InputMode
 from .key_binding import Registry
 from .key_bindings.emacs import emacs_bindings
-from .key_bindings.vi import vi_bindings
 from .line import Line, Callbacks
 from .prompt import Prompt
 from .renderer import Renderer
@@ -43,7 +42,7 @@ import weakref
 
 __all__ = (
         'AbortAction',
-        'CommandLine',
+        'CommandLineInterface',
 )
 
 class AbortAction:
@@ -64,7 +63,7 @@ class Abort(Exception):
     pass
 
 
-class CommandLine(object):
+class CommandLineInterface(object):
     """
     Wrapper around all the other classes, tying everything together.
     """
@@ -89,7 +88,7 @@ class CommandLine(object):
 
     #: `InputStreamHandler` class for the keybindings.
     key_bindings_factories = [ emacs_bindings ]
-    default_input_mode = InputMode.VI_INSERT
+    default_input_mode = InputMode.INSERT
 
     #: `Renderer` class.
     renderer_factory = Renderer
@@ -110,6 +109,10 @@ class CommandLine(object):
         self.stdout = stdout or sys.stdout
 
         #: The `Line` instance.
+                    # TODO: construct the `Line` object outside of this constructor.
+                    #       A commandline should be able to have several `Line` instances.
+                    #       e.g. one for the incremental-search input, one for
+                    #       another document, etc...
         self.line = self.line_factory(
                                 code_factory=self.code_factory,
                                 history_factory=self.history_factory,
@@ -117,7 +120,7 @@ class CommandLine(object):
 
         key_registry = Registry()
         for kb in self.key_bindings_factories:
-            kb(key_registry, self.line)
+            kb(key_registry, weakref.ref(self))
 
         #: The `InputProcessor` instance.
         self.input_processor = InputProcessor(key_registry, default_input_mode=self.default_input_mode)
@@ -272,7 +275,7 @@ class CommandLine(object):
             # Trigger read_start.
             self.on_read_input_start()
 
-            with raw_mode(self.stdin):
+            with raw_mode(self.stdin.fileno()):
                 self._redraw()
 
                 # When the window size changes, we do a redraw request call.
