@@ -2,9 +2,9 @@
 
 ::
 
-    from prompt_toolkit.contrib.python_import import PythonCommandLine
+    from prompt_toolkit.contrib.python_import import PythonCommandLineInterface
 
-    cli = PythonCommandLine()
+    cli = PythonCommandLineInterface()
     cli.read_input()
 """
 from __future__ import unicode_literals
@@ -13,15 +13,17 @@ from pygments.lexers import PythonLexer
 from pygments.style import Style
 from pygments.token import Keyword, Operator, Number, Name, Error, Comment, Token
 
-from prompt_toolkit import CommandLine
+from prompt_toolkit import CommandLineInterface
 from prompt_toolkit.code import Completion, Code, ValidationError
 from prompt_toolkit.enums import InputMode
 from prompt_toolkit.history import FileHistory, History
-from prompt_toolkit.key_bindings.vi import vi_bindings
 from prompt_toolkit.key_bindings.emacs import emacs_bindings
+from prompt_toolkit.key_bindings.vi import vi_bindings
+from prompt_toolkit.keys import Keys
 from prompt_toolkit.line import Line
 from prompt_toolkit.prompt import Prompt, TokenList, BracketsMismatchProcessor, PopupCompletionMenu, HorizontalCompletionMenu
-from prompt_toolkit.keys import Key
+from prompt_toolkit.renderer import Char
+from prompt_toolkit.selection import SelectionType
 
 import jedi
 import platform
@@ -30,7 +32,7 @@ import sys
 
 
 __all__ = (
-    'PythonCommandLine',
+    'PythonCommandLineInterface',
     'AutoCompletionStyle',
 )
 
@@ -72,6 +74,8 @@ class PythonStyle(Style):
         Token.IncrementalSearchMatch:         '#ffffff bg:#4444aa',
         Token.IncrementalSearchMatch.Current: '#ffffff bg:#44aa44',
 
+        Token.SelectedText:            '#ffffff bg:#6666aa',
+
         # Signature highlighting.
         Token.Signature:               '#888888',
         Token.Signature.Operator:      'bold #888888',
@@ -108,7 +112,7 @@ class PythonStyle(Style):
         Token.HorizontalMenu.Arrow:                   'bold #888888',
 
         # Grayed
-        Token.Aborted:                 '#aaaaaa',
+        Token.Aborted:                 '#888888',
 
         Token.ValidationError:         'bg:#aa0000 #ffffff',
 
@@ -116,14 +120,41 @@ class PythonStyle(Style):
         Token.Leftmargin.Tilde:   '#888888',
     }
 
+def _has_unclosed_brackets(text):
+    """
+    Starting at the end of the string. If we find an opening bracket
+    for which we didn't had a closing one yet, return True.
+    """
+    stack = []
 
-def python_bindings(registry, line):
+    # Ignore braces inside strings
+    text = re.sub(r'''('[^']*'|"[^"]*")''', '', text) # XXX: handle escaped quotes.!
+
+    for c in reversed(text):
+        if c in '])}':
+            stack.append(c)
+
+        elif c in '[({':
+            if stack:
+                if ((c == '[' and stack[-1] == ']') or
+                    (c == '{' and stack[-1] == '}') or
+                    (c == '(' and stack[-1] == ')')):
+                    stack.pop()
+            else:
+                # Opening bracket for which we didn't had a closing one.
+                return True
+
+    return False
+
+
+def python_bindings(registry, cli_ref):
     """
     Custom key bindings.
     """
+    line = cli_ref().line
     handle = registry.add_binding
 
-    @handle(Key.F6)
+    @handle(Keys.F6)
     def _(event):
         """
         Enable/Disable paste mode.
@@ -132,14 +163,15 @@ def python_bindings(registry, line):
         if line.paste_mode:
             line.is_multiline = True
 
-    @handle(Key.F7)
+    @handle(Keys.F7)
     def _(event):
         """
         Enable/Disable multiline mode.
         """
         line.is_multiline = not line.is_multiline
 
-    @handle(Key.Tab)
+    @handle(Keys.Tab, in_mode=InputMode.INSERT)
+    @handle(Keys.Tab, in_mode=InputMode.COMPLETE)
     def _(event):
         # When the 'tab' key is pressed with only whitespace character before the
         # cursor, do autocompletion. Otherwise, insert indentation.
@@ -148,14 +180,20 @@ def python_bindings(registry, line):
             line.insert_text('    ')
         else:
             line.complete_next()
-            event.input_processor.input_mode = InputMode.COMPLETE
+            if event.input_processor.input_mode != InputMode.COMPLETE:
+                event.input_processor.push_input_mode(InputMode.COMPLETE)
 
-    @handle(Key.BackTab, in_mode=InputMode.COMPLETE)
+    @handle(Keys.BackTab, in_mode=InputMode.INSERT)
+    @handle(Keys.BackTab, in_mode=InputMode.COMPLETE)
     def _(event):
         line.complete_previous()
 
-    @handle(Key.ControlJ) # TODO XXX: !!!!
-    @handle(Key.ControlM)
+        if event.input_processor.input_mode != InputMode.COMPLETE:
+            event.input_processor.push_input_mode(InputMode.COMPLETE)
+            line.complete_previous()
+
+    @handle(Keys.ControlJ, in_mode=InputMode.INSERT)
+    @handle(Keys.ControlM, in_mode=InputMode.INSERT)
     def _(event):
         _auto_enable_multiline()
 
@@ -176,9 +214,9 @@ def python_bindings(registry, line):
         cursor_at_the_end = line.document.is_cursor_at_the_end
 
         # If we just typed a colon, or still have open brackets, always insert a real newline.
-        if cursor_at_the_end and (line._colon_before_cursor() or
-                                  line._has_unclosed_brackets() or
-                                  line._starting_with_at()):
+        if cursor_at_the_end and (line.document.text_before_cursor.rstrip()[-1:] == ':' or
+                                  _has_unclosed_brackets(line.document.text_before_cursor) or
+                                  line.text.startswith('@')):
             line.is_multiline = True
 
         # If the character before the cursor is a backslash (line continuation
@@ -207,41 +245,9 @@ class PythonLine(Line):
         # Code signatures. (This is set asynchronously after a timeout.)
         self.signatures = []
 
-    def _text_changed(self):
+    def text_changed(self):
         self.is_multiline = '\n' in self.text
 
-    def _colon_before_cursor(self):
-        return self.document.text_before_cursor[-1:] == ':'
-
-    def _starting_with_at(self):
-        """ True when the line starts with a decorator. """
-        return self.text.startswith('@')
-
-    def _has_unclosed_brackets(self):
-        """ Starting at the end of the string. If we find an opening bracket
-        for which we didn't had a closing one yet, return True. """
-        text = self.document.text_before_cursor
-        stack = []
-
-        # Ignore braces inside strings
-        text = re.sub(r'''('[^']*'|"[^"]*")''', '', text) # XXX: handle escaped quotes.!
-
-        for c in reversed(text):
-            if c in '])}':
-                stack.append(c)
-
-            elif c in '[({':
-                if stack:
-                    if ((c == '[' and stack[-1] == ']') or
-                        (c == '{' and stack[-1] == '}') or
-                        (c == '(' and stack[-1] == ')')):
-                        stack.pop()
-                else:
-                    # Opening bracket for which we didn't had a closing one.
-                    return True
-
-        return False
-
     def newline(self):
         r"""
         Insert \n at the cursor position. Also add necessary padding.
@@ -267,45 +273,12 @@ class PythonLine(Line):
                 for x in range(4):
                     insert_text(' ')
 
-    def cursor_left(self):
-        """
-        When moving the cursor left in the left indentation margin, move four
-        spaces at a time.
-        """
-        before_cursor = self.document.current_line_before_cursor
-
-        if not self.paste_mode and before_cursor.isspace():
-            count = 1 + (len(before_cursor) - 1) % 4
-        else:
-            count = 1
-
-        for i in range(count):
-            super(PythonLine, self).cursor_left()
-
-    def cursor_right(self):
-        """
-        When moving the cursor right in the left indentation margin, move four
-        spaces at a time.
-        """
-        before_cursor = self.document.current_line_before_cursor
-        after_cursor = self.document.current_line_after_cursor
-
-        # Count space characters, after the cursor.
-        after_cursor_space_count = len(after_cursor) - len(after_cursor.lstrip())
-
-        if (not self.paste_mode and
-                    (not before_cursor or before_cursor.isspace()) and after_cursor_space_count):
-            count = min(4, after_cursor_space_count)
-        else:
-            count = 1
-
-        for i in range(count):
-            super(PythonLine, self).cursor_right()
-
 
 class PythonPrompt(Prompt):
     input_processors = [ BracketsMismatchProcessor() ]
 
+    min_height = 7
+
     @property
     def tokens_before_input(self):
         return [(Token.Prompt, 'In [%s]: ' % self.commandline.current_statement_index)]
@@ -325,7 +298,7 @@ class PythonPrompt(Prompt):
         """
         When inside functions, show signature.
         """
-        if self.commandline.input_processor.input_mode == InputMode.INCREMENTAL_SEARCH:
+        if self.commandline.input_processor.input_mode == InputMode.INCREMENTAL_SEARCH and self.line.isearch_state:
             screen.write_highlighted(list(self.isearch_prompt))
         elif self.commandline.input_processor.arg is not None:
             screen.write_highlighted(list(self.arg_prompt))
@@ -387,28 +360,29 @@ class PythonPrompt(Prompt):
 
     def _get_bottom_position(self, screen, last_screen_height):
         # Draw the menu at the bottom position.
-        return max(screen.current_height, last_screen_height - 2) - 1
+        return max(self.min_height - 2, screen.current_height, last_screen_height - 2) - 1
 
     def _print_tildes(self, screen):
         """
         Print tildes in the left margin between the last input line and the
-        toolbars. (This is what Vi-users like.)
+        toolbars.
         """
-        if self.commandline.vi_mode:
-            last_char_y, last_char_x = screen._cursor_mappings[len(self.line.document.text)]
+        last_char_y, last_char_x = screen._cursor_mappings[len(self.line.document.text)]
 
-            # Fill space with tildes
-            for y in range(last_char_y + 1, screen.current_height - 2):
-                screen.write_at_pos(y, 1, '~', Token.Leftmargin.Tilde)
+        # Fill space with tildes
+        for y in range(last_char_y + 1, screen.current_height - 2):
+            screen.write_at_pos(y, 1, Char('~', Token.Leftmargin.Tilde))
 
     def write_to_screen(self, screen, last_screen_height, accept=False, abort=False):
         self.write_before_input(screen)
-        self.write_input(screen)
+        self.write_input(screen, highlight=not (accept or abort))
 
         if not (accept or abort):
-            self.write_menus(screen)
-
             y = self._get_bottom_position(screen, last_screen_height)
+            self.write_menus(screen) # XXX: menu should be able to cover the second toolbar.
+            y2 = self._get_bottom_position(screen, last_screen_height)
+            y = max(y, y2 - 1)
+
 
             screen._y, screen._x = y + 1, 0
             self.write_second_toolbar(screen)
@@ -432,18 +406,25 @@ class PythonPrompt(Prompt):
             append((TB.Mode, '(SEARCH)'))
             append((TB, '   '))
         elif self.commandline.vi_mode:
-            if mode == InputMode.VI_NAVIGATION:
-                append((TB.Mode, '(NAV)'))
-                append((TB, '      '))
-            elif mode == InputMode.VI_INSERT:
+            if mode == InputMode.INSERT:
                 append((TB.Mode, '(INSERT)'))
                 append((TB, '   '))
+            elif mode == InputMode.VI_NAVIGATION:
+                append((TB.Mode, '(NAV)'))
+                append((TB, '      '))
             elif mode == InputMode.VI_REPLACE:
                 append((TB.Mode, '(REPLACE)'))
                 append((TB, '  '))
             elif mode == InputMode.COMPLETE:
                 append((TB.Mode, '(COMPLETE)'))
                 append((TB, ' '))
+            elif mode == InputMode.SELECTION:
+                if self.line.selection_state.type == SelectionType.LINES:
+                    append((TB.Mode, '(VISUAL LINE)'))
+                    append((TB, ' '))
+                elif self.line.selection_state.type == SelectionType.CHARACTERS:
+                    append((TB.Mode, '(VISUAL)'))
+                    append((TB, ' '))
 
 #            if self.commandline._input_processor.is_recording_macro:
 #                append((TB.Mode, 'recording'))
@@ -459,6 +440,9 @@ class PythonPrompt(Prompt):
         # Shortcuts.
         if mode == InputMode.INCREMENTAL_SEARCH:
             append((TB, '[Ctrl-G] Cancel search [Enter] Go to this position.'))
+        elif self.line.selection_state:
+            # Emacs cut/copy keys.
+            append((TB, '[Ctrl-W] Cut [Meta-W] Copy [Ctrl-Y] Paste [Ctrl-G] Cancel'))
         else:
             if self.line.paste_mode:
                 append((TB.On, '[F6] Paste mode (on)  '))
@@ -534,6 +518,9 @@ class PythonCode(Code):
             # gives `None` as offset in case of '4=4' as input. (Looks like
             # fixed in Python 3.)
             raise ValidationError(e.lineno - 1, (e.offset or 1) - 1, 'Syntax Error')
+        except TypeError as e:
+            # e.g. "compile() expected string without null bytes"
+            raise ValidationError(0, 0, str(e))
 
     def _get_jedi_script(self):
         try:
@@ -560,7 +547,7 @@ class PythonCode(Code):
                 yield PythonCompletion(c.name, len(c.complete) - len(c.name), c)
 
 
-class PythonCommandLine(CommandLine):
+class PythonCommandLineInterface(CommandLineInterface):
     line_factory = PythonLine
     prompt_factory = PythonPrompt
 
@@ -568,7 +555,7 @@ class PythonCommandLine(CommandLine):
                     style=PythonStyle, autocompletion_style=AutoCompletionStyle.POPUP_MENU):
 
         self.globals = globals or {}
-        self.locals = locals or {}
+        self.locals = locals or globals
         self.history_filename = history_filename
         self.style = style
         self.autocompletion_style = autocompletion_style
@@ -579,7 +566,7 @@ class PythonCommandLine(CommandLine):
         #: Incremeting integer counting the current statement.
         self.current_statement_index = 1
 
-        super(PythonCommandLine, self).__init__(stdin=stdin, stdout=stdout)
+        super(PythonCommandLineInterface, self).__init__(stdin=stdin, stdout=stdout)
 
     def history_factory(self):
         if self.history_filename:
