@@ -7,9 +7,11 @@ The `InputProcessor` will according to the implemented keybindings call the
 correct callbacks when new key presses are feed through `feed_key`.
 """
 from __future__ import unicode_literals
-from .keys import Key
+from .keys import Keys
 from .enums import InputMode
 
+import weakref
+
 __all__ = (
     'Registry',
     'InputProcessor',
@@ -17,13 +19,13 @@ __all__ = (
 
 
 class InputProcessor(object):
-    def __init__(self, registry, default_input_mode=InputMode.INPUT):
+    def __init__(self, registry, default_input_mode=InputMode.INSERT):
         self._registry = registry
         self.default_input_mode = default_input_mode
 
     def reset(self):
         self._previous_key_sequence = None
-        self.input_mode = self.default_input_mode
+        self._input_mode_stack = [self.default_input_mode]
 
         self._process_coroutine = self._process()
         self._process_coroutine.send(None)
@@ -32,6 +34,33 @@ class InputProcessor(object):
         #: https://www.gnu.org/software/bash/manual/html_node/Readline-Arguments.html
         self.arg = None
 
+    @property
+    def input_mode(self):
+        """
+        Current :class:`~prompt_toolkit.enums.InputMode`
+        """
+        return self._input_mode_stack[-1]
+
+    @input_mode.setter
+    def input_mode(self, value):
+        self._input_mode_stack[-1] = value
+
+    def push_input_mode(self, value):
+        """
+        Push new :class:`~prompt_toolkit.enums.InputMode` on the stack.
+        """
+        self._input_mode_stack.append(value)
+
+    def pop_input_mode(self):
+        """
+        Push new :class:`~prompt_toolkit.enums.InputMode` on the stack.
+        """
+        # You can't pop the last item.
+        if len(self._input_mode_stack) > 1:
+            return self._input_mode_stack.pop()
+        else:
+            raise IndexError('Cannot pop the last InputMode.')
+
     def _process(self):
         """
         Coroutine for the processing of key presses.
@@ -59,11 +88,11 @@ class InputProcessor(object):
                 key_sequence = processing + [key_press]
 
                 # Matches
-                exact_matches_with_mode = [(k, b) for k, b in options if k == (key_press,) and b.input_mode ]
-                exact_matches = [(k, b) for k, b in options if k == (key_press,)]
-                prefix_matches = [(k, b) for k, b in options if k[0] == key_press and len(k) > 1]
-                any_matches_with_mode = [(k, b) for k, b in options if k == (Key.Any,) and b.input_mode ]
-                any_matches = [(k, b) for k, b in options if k == (Key.Any,)]
+                exact_matches_with_mode = [(k, b) for k, b in options if k == (key_press.key,) and b.input_mode ]
+                exact_matches = [(k, b) for k, b in options if k == (key_press.key,)]
+                prefix_matches = [(k, b) for k, b in options if k[0] == key_press.key and len(k) > 1]
+                any_matches_with_mode = [(k, b) for k, b in options if k == (Keys.Any,) and b.input_mode ]
+                any_matches = [(k, b) for k, b in options if k == (Keys.Any,)]
 
                 # If we have an exact match, call handler.
                 # (An exact match that specifies the input mode will always get higher priority.)
@@ -76,11 +105,11 @@ class InputProcessor(object):
                     options = [(k[1:], b) for k, b in prefix_matches]
                     processing.append(key_press)
 
-                # If we have `Key.Any`, that should catch all keys (If there is
+                # If we have `Keys.Any`, that should catch all keys (If there is
                 # no other match) and pass the pressed key to the handler.
                 elif any_matches_with_mode or any_matches:
                     self._call_handler((any_matches_with_mode or any_matches)[0][1],
-                                    key_sequence=key_sequence, data=key_press)
+                                    key_sequence=key_sequence, key=key_press)
                     break
 
                 # An 'invalid' sequence, ignore the first key_press and start
@@ -89,11 +118,15 @@ class InputProcessor(object):
                     buffer = (processing + buffer)[1:]
                     break
 
-    def _call_handler(self, handler, data=None, key_sequence=None):
+    def feed_key(self, key_press):
+        self._process_coroutine.send(key_press)
+
+    def _call_handler(self, handler, key=None, key_sequence=None):
         arg = self.arg
         self.arg = None
 
-        event = Event(self, arg=arg, data=data, previous_key_sequence=self._previous_key_sequence)
+        event = Event(weakref.ref(self), arg=arg, data=(key and key.data),
+                        previous_key_sequence=self._previous_key_sequence)
         handler.call(event)
 
         for h in self._registry.after_handler_callbacks:
@@ -101,18 +134,19 @@ class InputProcessor(object):
 
         self._previous_key_sequence = key_sequence
 
-    def feed_key(self, key_press):
-        self._process_coroutine.send(key_press)
-
 
 class Event(object):
-    def __init__(self, input_processor, arg=None, data=None, key_sequence=None, previous_key_sequence=None):
-        self.input_processor = input_processor
+    def __init__(self, input_processor_ref, arg=None, data=None, key_sequence=None, previous_key_sequence=None):
+        self._input_processor_ref = input_processor_ref
         self.data = data
         self.key_sequence = key_sequence
         self.previous_key_sequence = previous_key_sequence
         self._arg = arg
 
+    @property
+    def input_processor(self):
+        return self._input_processor_ref()
+
     @property
     def arg(self):
         return self._arg or 1
@@ -153,6 +187,8 @@ class _Binding(object):
     def call(self, event):
         return self._callable(event)
 
+    def __repr__(self):
+        return '_Binding(keys=%r, callable=%r)' % (self.keys, self._callable)
 
 class Registry(object):
     """
