@@ -1,7 +1,8 @@
 from __future__ import unicode_literals
 from ..line import ClipboardData, ClipboardDataType
+from ..selection import SelectionType
 from ..enums import IncrementalSearchDirection, InputMode
-from ..keys import Key
+from ..keys import Keys
 
 from .basic import basic_bindings
 from .utils import create_handle_decorator
@@ -26,14 +27,15 @@ class CursorRegion(object):
             return self.end, self.start
 
 
-def vi_bindings(registry, line):
+def vi_bindings(registry, cli_ref):
     """
     Vi extensions.
 
     # Overview of Readline Vi commands:
     # http://www.catonmat.net/download/bash-vi-editing-mode-cheat-sheet.pdf
     """
-    basic_bindings(registry, line)
+    basic_bindings(registry, cli_ref)
+    line = cli_ref().line
     handle = create_handle_decorator(registry, line)
 
     _last_character_find = [None] # (char, backwards) tuple
@@ -56,48 +58,54 @@ def vi_bindings(registry, line):
                 len(line.document.current_line) > 0):
             line.cursor_position -= 1
 
-    @handle(Key.Escape)
+    @handle(Keys.Escape)
     def _(event):
         """
         Escape goes to vi navigation mode.
         """
-        event.input_processor.input_mode = InputMode.VI_NAVIGATION
+        if event.input_processor.input_mode == InputMode.SELECTION:
+            line.exit_selection()
+            event.input_processor.pop_input_mode()
+        else:
+            event.input_processor.input_mode = InputMode.VI_NAVIGATION
 
-    @handle(Key.Backspace, in_mode=InputMode.VI_NAVIGATION)
+    @handle(Keys.Backspace, in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         """
         In navigation-mode, move cursor.
         """
         line.cursor_position += line.document.get_cursor_left_position(count=event.arg)
 
-    #@handle(Key.ControlV, Key.Any, in_mode=InputMode.VI_INSERT)
-    #def _(event):
-    #    """
-    #    Insert a character literally (quoted insert).
-    #    """
-    #    line.insert_text(event.data, overwrite=False)
+    @handle(Keys.ControlV, Keys.Any, in_mode=InputMode.INSERT)
+    def _(event):
+        """
+        Insert a character literally (quoted insert).
+        """
+        line.insert_text(event.data, overwrite=False)
 
-    @handle(Key.ControlN)
+    @handle(Keys.ControlN, in_mode=InputMode.INSERT)
+    @handle(Keys.ControlN, in_mode=InputMode.COMPLETE)
     def _(event):
         line.complete_next()
 
         # Switch only to the 'complete' input mode if there really was a
         # completion found.
-        if line.complete_state:
-            event.input_processor.input_mode = InputMode.COMPLETE
+        if line.complete_state and event.input_processor.input_mode != InputMode.COMPLETE:
+            event.input_processor.push_input_mode(InputMode.COMPLETE)
 
-    @handle(Key.ControlP)
+    @handle(Keys.ControlP, in_mode=InputMode.INSERT)
+    @handle(Keys.ControlP, in_mode=InputMode.COMPLETE)
     def _(event):
         line.complete_previous()
 
         # Switch only to the 'complete' input mode if there really was a
         # completion found.
-        if line.complete_state:
-            event.input_processor.input_mode = InputMode.COMPLETE
+        if line.complete_state and event.input_processor.input_mode != InputMode.COMPLETE:
+            event.input_processor.push_input_mode(InputMode.COMPLETE)
 
 
-    @handle(Key.ControlJ, in_mode=InputMode.VI_NAVIGATION)
-    @handle(Key.ControlM, in_mode=InputMode.VI_NAVIGATION)
+    @handle(Keys.ControlJ, in_mode=InputMode.VI_NAVIGATION)
+    @handle(Keys.ControlM, in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         """
         In navigation mode, pressing enter will always return the input.
@@ -111,12 +119,12 @@ def vi_bindings(registry, line):
     @handle('a', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         line.cursor_position += line.document.get_cursor_right_position()
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('A', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         line.cursor_position += line.document.get_end_of_line_position()
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('C', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -128,7 +136,7 @@ def vi_bindings(registry, line):
         if deleted:
             data = ClipboardData(deleted)
             line.set_clipboard(data)
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('c', 'c', in_mode=InputMode.VI_NAVIGATION)
     @handle('S', in_mode=InputMode.VI_NAVIGATION)
@@ -143,7 +151,7 @@ def vi_bindings(registry, line):
         # But we delete after the whitespace
         line.cursor_position += line.document.get_start_of_line_position(after_whitespace=True)
         line.delete(count=line.document.get_end_of_line_position())
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('D', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -184,11 +192,11 @@ def vi_bindings(registry, line):
 
     @handle('i', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('I', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
         line.cursor_position += line.document.get_start_of_line_position(after_whitespace=True)
 
     @handle('J', in_mode=InputMode.VI_NAVIGATION)
@@ -233,7 +241,7 @@ def vi_bindings(registry, line):
         for i in range(event.arg):
             line.paste_from_clipboard(before=True)
 
-    @handle('r', Key.Any, in_mode=InputMode.VI_NAVIGATION)
+    @handle('r', Keys.Any, in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         """
         Replace single character under cursor
@@ -255,7 +263,7 @@ def vi_bindings(registry, line):
         """
         data = ClipboardData(''.join(line.delete() for i in range(event.arg)))
         line.set_clipboard(data)
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('u', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -266,6 +274,22 @@ def vi_bindings(registry, line):
     def _(event):
         line.open_in_editor()
 
+    # @handle('v', in_mode=InputMode.VI_NAVIGATION)
+    # def _(event):
+    #     """
+    #     Start characters selection.
+    #     """
+    #     line.start_selection(selection_type=SelectionType.CHARACTERS)
+    #     event.input_processor.push_input_mode(InputMode.SELECTION)
+
+    @handle('V', in_mode=InputMode.VI_NAVIGATION)
+    def _(event):
+        """
+        Start lines selection.
+        """
+        line.start_selection(selection_type=SelectionType.LINES)
+        event.input_processor.push_input_mode(InputMode.SELECTION)
+
     @handle('x', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         """
@@ -274,6 +298,27 @@ def vi_bindings(registry, line):
         data = ClipboardData(line.delete(count=event.arg))
         line.set_clipboard(data)
 
+    @handle('x', in_mode=InputMode.SELECTION)
+    @handle('d', 'd', in_mode=InputMode.SELECTION)
+    def _(event):
+        """
+        Cut selection.
+        """
+        selection_type = line.selection_state.type
+        deleted = line.cut_selection()
+        line.set_clipboard(ClipboardData(deleted, selection_type))
+        event.input_processor.pop_input_mode()
+
+    @handle('y', in_mode=InputMode.SELECTION)
+    def _(event):
+        """
+        Copy selection.
+        """
+        selection_type = line.selection_state.type
+        deleted = line.copy_selection()
+        line.set_clipboard(ClipboardData(deleted, selection_type))
+        event.input_processor.pop_input_mode()
+
     @handle('X', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         data = line.delete_before_cursor()
@@ -305,30 +350,6 @@ def vi_bindings(registry, line):
         line.cursor_position += line.document.get_cursor_up_position(count=event.arg)
         line.cursor_position += line.document.get_start_of_line_position(after_whitespace=True)
 
-    @handle('{', in_mode=InputMode.VI_NAVIGATION)
-    def _(event):
-        """
-        Move to previous blank-line separated section.
-        """
-        for i in range(event.arg):
-            index = line.document.find_previous_matching_line(
-                            lambda text: not text or text.isspace())
-
-            if index is not None:
-                line.cursor_position += line.document.get_cursor_up_position(count=index)
-
-    @handle('}', in_mode=InputMode.VI_NAVIGATION)
-    def _(event):
-        """
-        Move to next blank-line separated section.
-        """
-        for i in range(event.arg):
-            index = line.document.find_next_matching_line(
-                            lambda text: not text or text.isspace())
-
-            if index is not None:
-                line.cursor_position += line.document.get_cursor_down_position(count=index)
-
     @handle('>', '>', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
         """
@@ -363,7 +384,7 @@ def vi_bindings(registry, line):
         Open line above and enter insertion mode
         """
         line.insert_line_above()
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('o', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -371,7 +392,7 @@ def vi_bindings(registry, line):
         Open line below and enter insertion mode
         """
         line.insert_line_below()
-        event.input_processor.input_mode = InputMode.VI_INSERT
+        event.input_processor.input_mode = InputMode.INSERT
 
     @handle('~', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -389,7 +410,7 @@ def vi_bindings(registry, line):
         Search history backward for a command matching string.
         """
         line.incremental_search(IncrementalSearchDirection.FORWARD)
-        event.input_processor.input_mode = InputMode.INCREMENTAL_SEARCH
+        event.input_processor.push_input_mode(InputMode.INCREMENTAL_SEARCH)
 
     @handle('?', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -397,7 +418,7 @@ def vi_bindings(registry, line):
         Search history forward for a command matching string.
         """
         line.incremental_search(IncrementalSearchDirection.BACKWARD)
-        event.input_processor.input_mode = InputMode.INCREMENTAL_SEARCH
+        event.input_processor.push_input_mode(InputMode.INCREMENTAL_SEARCH)
 
     @handle('#', in_mode=InputMode.VI_NAVIGATION)
     def _(event):
@@ -436,6 +457,7 @@ def vi_bindings(registry, line):
         def decorator(func):
             if not no_move_handler:
                 @handle(*keys, in_mode=InputMode.VI_NAVIGATION)
+                @handle(*keys, in_mode=InputMode.SELECTION)
                 def move(event):
                     """ Create move handler. """
                     region = func(event)
@@ -475,7 +497,7 @@ def vi_bindings(registry, line):
 
                     # Only go back to insert mode in case of 'change'.
                     if not delete_only:
-                        event.input_processor.input_mode = InputMode.VI_INSERT
+                        event.input_processor.input_mode = InputMode.INSERT
 
             create(True)
             create(False)
@@ -515,9 +537,12 @@ def vi_bindings(registry, line):
         """ 'c^', 'd^' and '^': Soft start of line, after whitespace. """
         return CursorRegion(line.document.get_start_of_line_position(after_whitespace=True))
 
-    @change_delete_move_yank_handler('0')
+    @change_delete_move_yank_handler('0', no_move_handler=True)
     def key_zero(event):
-        """ 'c0', 'd0' and '0': Hard start of line, before whitespace. """
+        """
+        'c0', 'd0': Hard start of line, before whitespace.
+        (The move '0' key is implemented elsewhere, because a '0' could also change the `arg`.)
+        """
         return CursorRegion(line.document.get_start_of_line_position(after_whitespace=False))
 
     def create_ci_ca_handles(ci_start, ci_end, inner):
@@ -542,7 +567,38 @@ def vi_bindings(registry, line):
                             ('[', ']'), ('<', '>'), ('{', '}'), ('(', ')') ]:
             create_ci_ca_handles(ci_start, ci_end, inner)
 
-    @change_delete_move_yank_handler('f', Key.Any)
+    @change_delete_move_yank_handler('{') # TODO: implement 'arg'
+    def _(event):
+        """
+        Move to previous blank-line separated section.
+        Implements '{', 'c{', 'd{', 'y{'
+        """
+        line_index = line.document.find_previous_matching_line(
+                        lambda text: not text or text.isspace())
+
+        if line_index:
+            index = line.document.get_cursor_up_position(count=-line_index)
+        else:
+            index = 0
+        return CursorRegion(index)
+
+    @change_delete_move_yank_handler('}') # TODO: implement 'arg'
+    def _(event):
+        """
+        Move to next blank-line separated section.
+        Implements '}', 'c}', 'd}', 'y}'
+        """
+        line_index = line.document.find_next_matching_line(
+                        lambda text: not text or text.isspace())
+
+        if line_index:
+            index = line.document.get_cursor_down_position(count=line_index)
+        else:
+            index = 0
+
+        return CursorRegion(index)
+
+    @change_delete_move_yank_handler('f', Keys.Any)
     def _(event):
         """
         Go to next occurance of character. Typing 'fx' will move the
@@ -552,7 +608,7 @@ def vi_bindings(registry, line):
         match = line.document.find(event.data, in_current_line=True, count=event.arg)
         return CursorRegion(match or 0)
 
-    @change_delete_move_yank_handler('F', Key.Any)
+    @change_delete_move_yank_handler('F', Keys.Any)
     def _(event):
         """
         Go to previous occurance of character. Typing 'Fx' will move the
@@ -561,7 +617,7 @@ def vi_bindings(registry, line):
         _last_character_find[0] = (event.data, True)
         return CursorRegion(line.document.find_backwards(event.data, in_current_line=True, count=event.arg) or 0)
 
-    @change_delete_move_yank_handler('t', Key.Any)
+    @change_delete_move_yank_handler('t', Keys.Any)
     def _(event):
         """
         Move right to the next occurance of c, then one char backward.
@@ -570,7 +626,7 @@ def vi_bindings(registry, line):
         match = line.document.find(event.data, in_current_line=True, count=event.arg)
         return CursorRegion(match - 1 if match else 0)
 
-    @change_delete_move_yank_handler('T', Key.Any)
+    @change_delete_move_yank_handler('T', Keys.Any)
     def _(event):
         """
         Move left to the previous occurance of c, then one char forward.
@@ -656,15 +712,18 @@ def vi_bindings(registry, line):
         # Move to the top of the input.
         return CursorRegion(line.document.home_position)
 
-    @handle(Key.Any, in_mode=InputMode.VI_NAVIGATION)
+    @handle(Keys.Any, in_mode=InputMode.VI_NAVIGATION)
+    @handle(Keys.Any, in_mode=InputMode.SELECTION)
     def _(event):
         """
         Always handle numberics in navigation mode as arg.
         """
-        if event.data in '123456789' or (event.arg and event.data == '0'):
+        if event.data in '123456789' or (event._arg and event.data == '0'):
             event.append_to_arg_count(event.data)
+        elif event.data == '0':
+            line.cursor_position += line.document.get_start_of_line_position(after_whitespace=False)
 
-    @handle(Key.Any, in_mode=InputMode.VI_REPLACE)
+    @handle(Keys.Any, in_mode=InputMode.VI_REPLACE)
     def _(event):
         """
         Insert data at cursor position.
