@@ -8,6 +8,7 @@ from .code import Code, ValidationError
 from .document import Document
 from .enums import IncrementalSearchDirection
 from .history import History
+from .selection import SelectionType, SelectionState
 
 import os
 import tempfile
@@ -153,6 +154,9 @@ class Line(object):
         # State of Incremental-search
         self.isearch_state = None
 
+        # State of the selection.
+        self.selection_state = None
+
         # State of complete browser
         self.complete_state = None # For interactive completion through Ctrl-N/Ctrl-P.
 
@@ -181,8 +185,9 @@ class Line(object):
         # Remove validation errors, complete state.
         self.validation_error = None
         self.complete_state = None
+        self.selection_state = None
 
-        self._text_changed()
+        self.text_changed()
 
     @property
     def cursor_position(self):
@@ -196,6 +201,9 @@ class Line(object):
         self.validation_error = None
         self.complete_state = None
 
+        # Note that the cursor position can change if we have a selection the
+        # new position of the cursor determines the end of the selection.
+
     @property
     def _working_index(self):
         return self.__working_index
@@ -203,11 +211,17 @@ class Line(object):
     @_working_index.setter
     def _working_index(self, value):
         self.__working_index = value
-        self._text_changed()
+
+        # Remove any validation errors and complete state.
+        self.validation_error = None
+        self.complete_state = None
+        self.selection_state = None
+
+        self.text_changed()
 
     ### End of <getters/setters>
 
-    def _text_changed(self):
+    def text_changed(self):
         """
         Not implemented. Override to capture when the current visible text
         changes.
@@ -259,7 +273,7 @@ class Line(object):
         position.
         """
         # TODO: this can be cached as long self.text does not change.
-        return Document(self.text, self.cursor_position)
+        return Document(self.text, self.cursor_position, selection=self.selection_state)
 
     def set_arg_prompt(self, arg): # XXX: Just make a property of this???
         """
@@ -287,21 +301,21 @@ class Line(object):
     def auto_up(self, count=1):
         """
         If we're not on the first line (of a multiline input) go a line up,
-        otherwise go back in history.
+        otherwise go back in history. (If nothing is selected.)
         """
         if self.document.cursor_position_row > 0:
             self.cursor_position += self.document.get_cursor_up_position(count=count)
-        else:
+        elif not self.selection_state:
             self.history_backward(count=count)
 
     def auto_down(self, count=1):
         """
         If we're not on the last line (of a multiline input) go a line down,
-        otherwise go forward in history.
+        otherwise go forward in history. (If nothing is selected.)
         """
         if self.document.cursor_position_row < self.document.line_count - 1:
             self.cursor_position += self.document.get_cursor_down_position(count=count)
-        else:
+        elif not self.selection_state:
             old_index = self._working_index
             self.history_forward(count=count)
 
@@ -473,6 +487,38 @@ class Line(object):
             self._working_index -= count
             self.cursor_position = len(self.text)
 
+    def start_selection(self, selection_type=SelectionType.CHARACTERS):
+        """
+        Take the current cursor position as the start of this selection.
+        """
+        self.selection_state = SelectionState(self.cursor_position, selection_type)
+
+    def copy_selection(self, _cut=False):
+        """
+        Copy selected text and return it as string.
+        """
+        if self.selection_state:
+            # Take start and end of selection
+            from_, to = self.document.selection_range()
+
+            copied_text = self.text[from_:to]
+
+            # If cutting, remove the text and set the new cursor position.
+            if _cut:
+                self.text = self.text[:from_] + self.text[to + 1:]
+                self.cursor_position = min(from_, to)
+
+            self.selection_state = None
+            return copied_text
+        else:
+            return ''
+
+    def cut_selection(self):
+        """
+        Delete the selected text and return it as string.
+        """
+        return self.copy_selection(_cut=True)
+
     def newline(self):
         self.insert_text('\n')
 
@@ -592,7 +638,7 @@ class Line(object):
 
     def return_input(self):
         """
-        Return the current line to the `CommandLine.read_input` call.
+        Return the current line to the `CommandLineInterface.read_input` call.
         """
         code = self.create_code_obj()
         text = self.text
@@ -688,6 +734,9 @@ class Line(object):
 
         self.isearch_state = None
 
+    def exit_selection(self):
+        self.selection_state = None
+
     def clear(self):
         """
         Clear screen, usually as a result of Ctrl-L.
