@@ -341,10 +341,12 @@ class Prompt(object):
         """
         Tokens for the prompt when we go in reverse-i-search mode.
         """
-        return self.isearch_composer(self.line.isearch_state).get_tokens()
+        if self.line.isearch_state:
+            return self.isearch_composer(self.line.isearch_state).get_tokens()
+        else:
+            return []
 
-    @property
-    def tokens_after_input(self):
+    def get_tokens_after_input(self):
         """
         List of (Token, text) tuples for after the inut.
         (This can be used to create a help text or a status line.)
@@ -364,11 +366,13 @@ class Prompt(object):
         length = len(prompt_text.rstrip())
         spaces = len(prompt_text) - length
 
-        yield (Token.Prompt.SecondLinePrefix, '.' * length)
-        yield (Token.Prompt.SecondLinePrefix, ' ' * spaces)
+        return [
+            (Token.Prompt.SecondLinePrefix, '.' * length),
+            (Token.Prompt.SecondLinePrefix, ' ' * spaces)
+        ]
 
     def create_left_input_margin(self, screen, row, is_new_line):
-        screen.write_highlighted(self.tokens_in_left_margin)
+        screen.write_highlighted(self.get_tokens_in_left_margin(row, is_new_line))
 
     def write_before_input(self, screen):
         screen.write_highlighted(self.tokens_before_input)
@@ -399,13 +403,24 @@ class Prompt(object):
                 highlighted_characters.update({
                     x: token for x in range(index, index + len(self.line.isearch_state.isearch_text)) })
 
+        # In case of selection, highlight all matches.
+        selection_range = self.line.document.selection_range()
+        if selection_range:
+            from_, to = selection_range
+
+            for i in range(from_, to):
+                highlighted_characters[i] = Token.SelectedText
+
         return highlighted_characters
 
-    def write_input(self, screen):
+    def write_input(self, screen, highlight=True):
         # Set second line prefix
         screen.set_left_margin(lambda row, is_new_line: self.create_left_input_margin(screen, row, is_new_line))
 
-        highlighted_characters = self.get_highlighted_characters()
+        if highlight:
+            highlighted_characters = self.get_highlighted_characters()
+        else:
+            highlighted_characters = {}
 
         index = 0
             # Note, we add the space character at the end, because that's where
@@ -430,7 +445,7 @@ class Prompt(object):
         """
         Write tokens after input.
         """
-        screen.write_highlighted(self.tokens_after_input)
+        screen.write_highlighted(self.get_tokens_after_input())
 
     def write_menus(self, screen):
         """
@@ -448,7 +463,7 @@ class Prompt(object):
         Render the prompt to a `Screen` instance.
         """
         self.write_before_input(screen)
-        self.write_input(screen)
+        self.write_input(screen, highlight=not (accept or abort))
 
         if not (accept or abort):
             self.write_after_input(screen)
