@@ -104,63 +104,60 @@ class TerminalCodes:
 
 
 class Char(object):
-    __slots__ = ('char', 'style', 'z_index')
+    __slots__ = ('char', 'token', 'z_index')
+
+    # If we end up having one of these special control sequences in the input string,
+    # we should display them as follows:
+    # Usually this happens after a "quoted insert".
+    display_mappings = {
+        '\x1b': '^[', # Escape
+        '\x01': '^A',
+        '\x02': '^B',
+        '\x03': '^C',
+        '\x04': '^D',
+        '\x05': '^E',
+        '\x06': '^F',
+        '\x07': '^G',
+        '\x08': '^H',
+        '\x09': '^I',
+        '\x0a': '^J',
+        '\x0b': '^K',
+        '\x0c': '^L',
+        '\x0d': '^M',
+        '\x0e': '^N',
+        '\x0f': '^O',
+        '\x10': '^P',
+        '\x11': '^Q',
+        '\x12': '^R',
+        '\x13': '^S',
+        '\x14': '^T',
+        '\x15': '^U',
+        '\x16': '^V',
+        '\x17': '^W',
+        '\x18': '^X',
+        '\x19': '^Y',
+        '\x1a': '^Z',
+        '\x00': '^ ', # Control space
+        '\x1c': '^\\',
+        '\x1d': '^]',
+        '\x1f': '^_',
+        '\x7f': '^?', # Control backspace
+    }
+
+    def __init__(self, char=' ', token=None, z_index=0):
+        # If this character has to be displayed otherwise, take that one.
+        char = self.display_mappings.get(char, char)
 
-    def __init__(self, char=' ', style=None, z_index=0):
         self.char = char
-        self.style = style # TODO: maybe we should still use `token` instead of
-                           #       `style` and use the actual style in the last step of the renderer.
+        self.token = token
         self.z_index = z_index
 
-    def output(self, last_char=None):
-        """ Return the output to write this character to the terminal. """
-        style = self.style
-
-        # If the last printed charact has the same styling, we don't have to
-        # output the style.
-        if last_char and last_char.style == self.style:
-            return self.char
-
-        if style:
-            e = EscapeSequence(
-                    fg=(_tf._color_index(style['color']) if style['color'] else None),
-                    bg=(_tf._color_index(style['bgcolor']) if style['bgcolor'] else None),
-                    bold=style.get('bold', False),
-                    underline=style.get('underline', False))
-
-            return ''.join([
-                    TerminalCodes.RESET_ATTRIBUTES,
-                    e.color_string(),
-                    self.char,
-                ])
-        else:
-            return ''.join([
-                TerminalCodes.RESET_ATTRIBUTES,
-                self.char
-                ])
-
     @property
     def width(self):
         # We use the `max(0, ...` because some non printable control
         # characters, like e.g. Ctrl-underscore get a -1 wcwidth value.
         # It can be possible that these characters end up in the input text.
-        return max(0, wcwidth(self.char))
-
-    @property
-    def _background(self):
-        if self.style:
-            return self.style.get('bgcolor', None)
-
-    def __eq__(self, other):
-        """
-        Test whether two `Char` instances are equal.
-        """
-        # Spaces are always equal, whathever their styling.
-        if self.char == other.char and self.char == ' ' and self._background == other._background:
-            return True
-
-        # We ignore z-index, that does not matter if things get painted.
-        return self.char == other.char and self.style == other.style
+        return max(0, sum([wcwidth(c) for c in self.char]))
 
 
 class Screen(object):
@@ -170,7 +167,7 @@ class Screen(object):
     :param style: Pygments style.
     :param grayed: True when all tokes should be replaced by `Token.Aborted`
     """
-    def __init__(self, style, size, grayed=False):
+    def __init__(self, size):
         self._buffer = defaultdict(lambda: defaultdict(Char))
         self._cursor_mappings = { } # Map `source_string_index` of input data to (row, col) screen output.
         self._x = 0
@@ -178,12 +175,9 @@ class Screen(object):
         self._max_y = 0
 
         self.size = size
-        self._style = style
-        self._grayed = grayed
         self._left_margin_func = lambda linenr, is_new_line: None
 
-        self._cursor_x = 0
-        self._cursor_y = 0
+        self.cursor_position = _Point(y=0, x=0)
         self._line_number = 1
 
     @property
@@ -194,7 +188,7 @@ class Screen(object):
             return max(self._buffer.keys()) + 1
 
     def get_cursor_position(self):
-        return self._cursor_y, self._cursor_x
+        return self.cursor_position
 
     def set_left_margin(self, func):
         """
@@ -210,7 +204,8 @@ class Screen(object):
         """
         assert len(char) == 1
 
-        char_width = wcwidth(char)
+        char_obj  = Char(char, token, z_index)
+        char_width = char_obj.width
 
         # In case there is no more place left at this line, go first to the
         # following line. (Also in case of double-width characters.)
@@ -225,12 +220,7 @@ class Screen(object):
             self._cursor_mappings[string_index] = insert_pos
 
         if set_cursor_position:
-            self._cursor_x = self._x
-            self._cursor_y = self._y
-
-        # If grayed, replace token
-        if self._grayed:
-            token = Token.Aborted
+            self.cursor_position = _Point(y=self._y, x=self._x)
 
         # Insertion of newline
         if char == '\n':
@@ -241,28 +231,22 @@ class Screen(object):
 
         # Insertion of a 'visible' character.
         else:
-            self.write_at_pos(self._y, self._x, char, token, z_index=z_index)
+            self.write_at_pos(self._y, self._x, char_obj)
 
-            # Move cursor position
+            # Move position
             self._x += char_width
 
         return insert_pos
 
-    def write_at_pos(self, y, x, char, token, z_index=0):
+    def write_at_pos(self, y, x, char_obj): #char, token, z_index=0):
         """
         Write character at position (y, x).
         (Truncate when character is outside margin.)
         """
-        # Get style
-        try:
-            style = self._style.style_for_token(token)
-        except KeyError:
-            style = None
-
         # Add char to buffer
         if y < self.size.columns:
-            if z_index >= self._buffer[y][x].z_index:
-                self._buffer[y][x] = Char(char=char, style=style, z_index=z_index)
+            if char_obj.z_index >= self._buffer[y][x].z_index:
+                self._buffer[y][x] = char_obj
 
     def write_highlighted_at_pos(self, y, x, data, z_index=0):
         """
@@ -271,7 +255,7 @@ class Screen(object):
         """
         for token, text in data:
             for c in text:
-                self.write_at_pos(y, x, c, token, z_index=z_index)
+                self.write_at_pos(y, x, Char(c, token, z_index=z_index))
                 x += wcwidth(c)
 
     def write_highlighted(self, data):
@@ -282,118 +266,170 @@ class Screen(object):
             for c in text:
                 self.write_char(c, token=token)
 
-    def output(self, current_pos, previous_screen=None, last_char=None, accept_or_abort=False):
-        """
-        Create diff of this screen with the previous screen.
-        """
-        last_char = [last_char] # nonlocal
-        result = []
-        write = result.append
 
-        def move_cursor(new):
-            if current_pos.x >= self.size.columns - 1:
-                write(TerminalCodes.CARRIAGE_RETURN)
-                write(TerminalCodes.CURSOR_FORWARD(new.x))
-            elif new.x < current_pos.x or current_pos.x >= self.size.columns - 1:
-                write(TerminalCodes.CURSOR_BACKWARD(current_pos.x - new.x))
-            elif new.x > current_pos.x:
-                write(TerminalCodes.CURSOR_FORWARD(new.x - current_pos.x))
+def output_screen_diff(screen, current_pos, previous_screen=None, last_char=None, accept_or_abort=False, style=None, grayed=False):
+    """
+    Create diff of this screen with the previous screen.
+    """
+    #: Remember the last printed character.
+    last_char = [last_char] # nonlocal
+    background_turned_on = [False] # Nonlocal
 
-            if new.y > current_pos.y:
-                # Use newlines instead of CURSOR_DOWN, because this meight add new lines.
-                # CURSOR_DOWN will never create new lines at the bottom.
-                # Also reset attributes, otherwise the newline could draw a
-                # background color.
-                write(TerminalCodes.RESET_ATTRIBUTES)
-                write(TerminalCodes.NEWLINE * (new.y - current_pos.y))
-                write(TerminalCodes.CURSOR_FORWARD(new.x))
-                last_char[0] = None # Forget last char after resetting attributes.
-            elif new.y < current_pos.y:
-                write(TerminalCodes.CURSOR_UP(current_pos.y - new.y))
+    #: Variable for capturing the output.
+    result = []
+    write = result.append
 
-            return new
+    def move_cursor(new):
+        if current_pos.x >= screen.size.columns - 1:
+            write(TerminalCodes.CARRIAGE_RETURN)
+            write(TerminalCodes.CURSOR_FORWARD(new.x))
+        elif new.x < current_pos.x or current_pos.x >= screen.size.columns - 1:
+            write(TerminalCodes.CURSOR_BACKWARD(current_pos.x - new.x))
+        elif new.x > current_pos.x:
+            write(TerminalCodes.CURSOR_FORWARD(new.x - current_pos.x))
 
-        # Disable autowrap
-        if not previous_screen:
-            write(TerminalCodes.DISABLE_AUTOWRAP)
+        if new.y > current_pos.y:
+            # Use newlines instead of CURSOR_DOWN, because this meight add new lines.
+            # CURSOR_DOWN will never create new lines at the bottom.
+            # Also reset attributes, otherwise the newline could draw a
+            # background color.
             write(TerminalCodes.RESET_ATTRIBUTES)
+            write(TerminalCodes.NEWLINE * (new.y - current_pos.y))
+            write(TerminalCodes.CURSOR_FORWARD(new.x))
+            last_char[0] = None # Forget last char after resetting attributes.
+        elif new.y < current_pos.y:
+            write(TerminalCodes.CURSOR_UP(current_pos.y - new.y))
 
-        # Get height of the screen.
-        # (current_height changes as we loop over _buffer, so remember the current value.)
-        current_height = self.current_height
+        return new
 
-        if previous_screen:
-            row_count = max(self.current_height, previous_screen.current_height)
+    def get_style_for_token(token, replace_if_grayed=True):
+        """
+        Get style
+        """
+        # If grayed, replace token
+        if grayed and replace_if_grayed:
+            token = Token.Aborted
+
+        try:
+            return style.style_for_token(token)
+        except KeyError:
+            return None
+
+    def chars_are_equal(new_char, old_char):
+        """
+        Test whether two `Char` instances are equal if printed.
+        """
+        new_token = Token.Aborted if grayed else new_char.token
+
+        # If they are both a space and have the same background color, we
+        # always consider them equal.
+        if new_char.char == ' ' and old_char.char == ' ':
+            style1 = get_style_for_token(new_token, replace_if_grayed=False)
+            style2 = get_style_for_token(old_char.token, replace_if_grayed=False)
+            return (style1 and style1.get('bgcolor')) == (style2 and style2.get('bgcolor'))
+
+        # We ignore z-index, that does not matter if things get painted.
+        return new_char.char == old_char.char and new_token == old_char.token
+
+    def output_char(char):
+        """
+        Write the output of this charact.r
+        """
+        # If the last printed character has the same token, it also has the
+        # same style, so we don't output it.
+        if last_char[0] and last_char[0].token == char.token:
+            write(char.char)
         else:
-            row_count = self.current_height
+            style = get_style_for_token(char.token)
 
-        # When the previous screen has a different width, redraw everything anyway.
-        if previous_screen and previous_screen.size != self.size:
-            current_pos = move_cursor(_Point(0, 0))
-            write(TerminalCodes.RESET_ATTRIBUTES)
-            write(TerminalCodes.ERASE_DOWN)
-            previous_screen = None
+            if style:
+                # Create new style and output.
+                fg = _tf._color_index(style['color']) if style['color'] else None
+                bg = _tf._color_index(style['bgcolor']) if style['bgcolor'] else None
 
-        # Loop over the rows.
-        c = 0
+                e = EscapeSequence(fg=fg, bg=bg,
+                        bold=style.get('bold', False),
+                        underline=style.get('underline', False))
 
-        for y, r in enumerate(range(0, row_count)):
-            new_row = self._buffer[r]
-            if previous_screen:
-                previous_row = previous_screen._buffer[r]
+                write(TerminalCodes.RESET_ATTRIBUTES)
+                write(e.color_string())
+                write(char.char)
 
-            if new_row:
-                new_max_line_len = max(new_row.keys())
+                # If we printed something with a background color, remember that.
+                background_turned_on[0] = bool(bg)
             else:
-                new_max_line_len = 0
-
-            if previous_screen:
-                if previous_row:
-                    previous_max_line_len = max(previous_row.keys())
-                else:
-                    previous_max_line_len = 0
-
-            # Loop over the columns.
-            c = 0
-            while c < new_max_line_len + 1:
-                char_width = (new_row[c].width or 1)
-
-                if not previous_screen or not (new_row[c] == previous_row[c]):
-                    current_pos = move_cursor(_Point(y=y, x=c))
-                    write(new_row[c].output(last_char[0]))
-                    last_char[0] = new_row[c]
-                    current_pos = current_pos._replace(x=current_pos.x + char_width)
-
-                c += char_width
-
-            # If the new line is shorter, trim it
-            if previous_screen and new_max_line_len < previous_max_line_len:
-                current_pos = move_cursor(_Point(y=y, x=new_max_line_len+1))
+                # Reset previous style and output.
                 write(TerminalCodes.RESET_ATTRIBUTES)
-                write(TerminalCodes.ERASE_END_OF_LINE)
-                last_char[0] = None # Forget last char after resetting attributes.
+                write(char.char)
 
-        # Move cursor:
-        if accept_or_abort:
-            current_pos = move_cursor(_Point(y=current_height, x=0))
-            write(TerminalCodes.CRLF)
-            write(TerminalCodes.ERASE_DOWN)
-        else:
-            cursor_y, cursor_x = self.get_cursor_position()
-            current_pos = move_cursor(_Point(y=cursor_y, x=cursor_x))
+        last_char[0] = char
 
-        if accept_or_abort:
+    # Disable autowrap
+    if not previous_screen:
+        write(TerminalCodes.DISABLE_AUTOWRAP)
+        write(TerminalCodes.RESET_ATTRIBUTES)
+
+    # When the previous screen has a different width, redraw everything anyway.
+    if not previous_screen or previous_screen.size != screen.size:
+        current_pos = move_cursor(_Point(0, 0))
+        write(TerminalCodes.RESET_ATTRIBUTES)
+        write(TerminalCodes.ERASE_DOWN)
+
+        previous_screen = Screen(screen.size)
+
+    # Get height of the screen.
+    # (current_height changes as we loop over _buffer, so remember the current value.)
+    current_height = screen.current_height
+
+    # Loop over the rows.
+    row_count = max(screen.current_height, previous_screen.current_height)
+    c = 0 # Column counter.
+
+    for y, r in enumerate(range(0, row_count)):
+        new_row = screen._buffer[r]
+        previous_row = previous_screen._buffer[r]
+
+        new_max_line_len = max(new_row.keys()) if new_row else 0
+        previous_max_line_len = max(previous_row.keys()) if previous_row else 0
+
+        # Loop over the columns.
+        c = 0
+        while c < new_max_line_len + 1:
+            char_width = (new_row[c].width or 1)
+
+            if not chars_are_equal(new_row[c], previous_row[c]):
+                current_pos = move_cursor(_Point(y=y, x=c))
+                output_char(new_row[c])
+                current_pos = current_pos._replace(x=current_pos.x + char_width)
+
+            c += char_width
+
+        # If the new line is shorter, trim it
+        if previous_screen and new_max_line_len < previous_max_line_len:
+            current_pos = move_cursor(_Point(y=y, x=new_max_line_len+1))
             write(TerminalCodes.RESET_ATTRIBUTES)
-            write(TerminalCodes.ENABLE_AUTOWRAP)
+            write(TerminalCodes.ERASE_END_OF_LINE)
+            last_char[0] = None # Forget last char after resetting attributes.
 
-        # If the last printed character has a background color, always reset.
-        # (Many terminals give weird artifacs on resize events when there is an
-        # active background color.)
-        if last_char[0] and last_char[0]._background:
-            write(TerminalCodes.RESET_ATTRIBUTES)
-            last_char[0] = None
+    # Move cursor:
+    if accept_or_abort:
+        current_pos = move_cursor(_Point(y=current_height, x=0))
+        write(TerminalCodes.ERASE_DOWN)
+    else:
+        current_pos = move_cursor(screen.get_cursor_position())
 
-        return ''.join(result), current_pos, last_char[0]
+    if accept_or_abort:
+        write(TerminalCodes.RESET_ATTRIBUTES)
+        write(TerminalCodes.ENABLE_AUTOWRAP)
+
+    # If the last printed character has a background color, always reset.
+    # (Many terminals give weird artifacs on resize events when there is an
+    # active background color.)
+    if background_turned_on[0]:
+        write(TerminalCodes.RESET_ATTRIBUTES)
+        last_char[0] = None
+
+    return ''.join(result), current_pos, last_char[0]
 
 
 class Renderer(object):
@@ -436,13 +472,14 @@ class Renderer(object):
 
     def render_to_str(self, abort=False, accept=False):
         # Generate the output of the new screen.
-        screen = self.screen_cls(style=self._style, size=self.get_size(), grayed=abort)
+        screen = self.screen_cls(size=self.get_size())
 
         self.prompt.write_to_screen(screen, self._last_screen.current_height if self._last_screen else 0,
                 abort=abort, accept=accept)
 
-        output, self._cursor_pos, self._last_char = screen.output(self._cursor_pos,
-                self._last_screen, self._last_char, accept or abort)
+        output, self._cursor_pos, self._last_char = output_screen_diff(screen, self._cursor_pos,
+                self._last_screen, self._last_char, accept or abort,
+                style=self._style, grayed=abort)
         self._last_screen = screen
 
         return output
